> module Core (module Core, Lab(..), ppLab, UnOp(..), BinOp(..), Prim(..))
> where
> import Data.List (sort)
> import Data.Set (union, difference, unions, Set, singleton)
> import qualified Data.Set as S
>
> import AbstractSyntax (Lab(..), ppLab, UnOp(..), BinOp(..), Prim(..))
> import qualified AbstractSyntax as AS
> import Message hiding (And)

> -- for printing core expressions
> import Printer hiding (ppPat, ppRow) 
> import Control.Monad.Reader
> import Options

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Expressions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Var   =  String
> type Con   =  String

> data Expr  
>   =  Bool     Bool                            -- truth value
>   |  If       Expr Expr Expr                  -- alternative
>   |  UnOp     UnOp Expr                       -- unary operator (on the naturals)
>   |  BinOp    Expr BinOp Expr                 -- binary operator (on the naturals)
>   |  Nat      Integer                         -- numeral
>   |  String   String                          -- string literal
>   |  Type                                     -- type argument marker
>   |  Var      Var                             -- variable
>   |  Let      Expr Expr                       -- local declaration
>   |  Fun      Pat  Expr                       -- abstraction
>   |  App      Expr Expr                       -- application
>   |  Rec      Pat  Expr                       -- recursion
>   |  Record   [Row Expr]                      -- record construction
>   |  Sel      Expr Lab                        -- record selection
>   |  Con      Con Expr                        -- constructor application
>   |  Case     Expr [(Pat, Expr)]              -- case analysis
>--   |  TyApp    Expr Type                     -- type application
>   |  Seq      Expr Expr                       -- sequence: e1 ; e2 ~= let _ = e1 in e2
>   |  Assign   Expr Expr                       -- assignment
>   |  Ref      Expr                            -- allocation
>   |  Deref    Expr                            -- dereferencing
>   |  LetCC    Var  Expr                       -- let current continuation
>   |  Continue Expr Expr                       -- apply a continuation
>   |  Throw    Expr                            -- throw an exception
>   |  Try      Expr [(Pat, Expr)]              -- catch an exception
>   |  Call     Prim Expr
>   |  While    Expr Expr                       -- while-loop

First-class environments.

>   |  Empty                                    -- empty environment
>   |  Val      Pat  Expr			-- binding 
>   |  Comma    Expr Expr                       -- union of two environments (comma operator)
>   |  Lookup   Expr Var			-- environment lookup

Objects and classes.

>   |  Class  Expr
>   |  Object Expr
>   |  New Expr
>   |  Method Var Expr
>   |  Append Expr Expr
>   |  Invoke Expr Lab Expr
>   |  Inherit Expr
>   |  InheritIn Var Expr Expr
>   |  Include Expr
>   |  IncludeIn Var Expr Expr

Arrays.

>   |  ArrFun Expr Var Expr
>   |  ArrList [Expr]
>   |  Subscript Expr Expr

Fr Evaluate, um Frames in Expr berfhren zu knnen und dann
konsistent ausgeben zu knnen:

> --  |  Hole
>      deriving (Show)

> unit       =  Record []

Records are sorted internally, but only \emph{after} evaluation
(so that the evaluation order is the textual order).

> append Empty e  =  e
> append e Empty  =  e
> append e1 e2    =  Append e1 e2

> comma Empty e  =  e
> comma e Empty  =  e
> comma e1 e2    =  Comma e1 e2

> instance Pretty Expr

> ppBool :: Bool -> P Doc
> ppBool False  =  keyword' "false"
> ppBool True   =  keyword' "true"


> pExpr :: Options -> Int -> Expr -> Doc
> pExpr o p e = runReader (ppExpr e) (defaultPState { pcolor = color o, pdebug = debug o, pprec = p })

> ppExpr :: Expr -> P Doc
> ppExpr (Bool b) 
>   =  ppBool b
> ppExpr (If e1 e2 e3)
>   =  do
>        pki <- keyword "if"
>        pkt <- keyword "then"
>        pke <- keyword "else"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        pe3 <- withPrec 0 $ ppExpr e3
>        return (group (pki <> 
>                       block 3 (pe1 <+> block 5 (pkt <> pe2) <+>
>                                        block 5 (pke <> pe3))))
> ppExpr (UnOp op e)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        return (group (pretty op <> text " " <> block 2 pe))
> ppExpr (BinOp e1 op e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <> text " " <> block 2 (pretty op <+> pe2)))
> ppExpr (Nat n) 
>   =  return (pretty n)
> ppExpr (String s) 
>   =  return (pretty s)
> ppExpr Type
>   =  do
>        pc <- ppCol astype
>        return (pc (text "<>"))
> ppExpr (Var x) 
>   =  return (text x)
> ppExpr (Let e1 e2)
>   =  do
>        pkl <- keyword' "let"
>        pki <- keyword' "in"
>        pke <- keyword' "end"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        return (group (pkl <+> block 4 pe1 <> 
>                       block 1 (sp <> pki <+> block 3 pe2) <+> pke))
> ppExpr (Fun p e)
>   =  parensPrec 10 $ do
>        pp <- withPrec 0 $ ppPat p
>        pe <- withPrec 0 $ ppExpr e
>        alignkey "fun" (pp <> text " =>" <+> pe)
> ppExpr (App e1 e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 10 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <> text " @" <+> pe2))
> ppExpr (Rec p e)
>   =  parensPrec 10 $ do
>        pp <- withPrec 0 $ ppPat p
>        pe <- withPrec 0 $ ppExpr e
>        alignkey "rec" (pp <> text " => " <+> pe)
> ppExpr (Sel e l)
>   =  do
>        pe <- withPrec 11 $ ppExpr e
>        return (pe <> text "." <> ppLab l)
> ppExpr (Record rs)
>   =  do
>        prs <- withPrec 0 $ mapM (ppRow ppExpr) rs
>        return (parens (group (intersperse (char ',' <> sp) prs)))
> ppExpr (Con c e) 
>   =  parensPrec 10 $ do
>        pc <- ppCon c
>        pe <- withPrec 11 $ ppExpr e
>        return (pc <+> pe)
> ppExpr (Case e rs)
>   =  do
>        pkc <- keyword "case"
>        pko <- keyword "of"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        prs <- withPrec 0 $ mapM ppRule rs
>        return (group (pkc <> pe <+> pko <>
>                       intersperse (sp <> text " | ") prs <+> pke))
> ppExpr (Seq e1 e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <> text ";" <+> pe2))
> ppExpr (Assign e1 e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <> text " :=" <+> pe2))
> ppExpr (Ref e)
>   =  do
>        pk <- keyword' "ref"
>        parensPrec 10 $ do
>          pe <- withPrec 11 $ ppExpr e
>          return (group (pk <+> pe))
> ppExpr (Deref e)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        return (group (text "!" <> pe))
> ppExpr (LetCC x e)
>   =  do
>        pkl <- keyword' "letcc"
>        pki <- keyword' "in"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pkl <+> block 6 (text x) <> 
>                       block 1 (sp <> pki <+> block 3 pe) <+> pke))
> ppExpr (Continue e1 e2)
>   =  do
>        pk <- keyword' "continue"
>        parensPrec 10 $ do
>          pe1 <- withPrec 11 $ ppExpr e1
>          pe2 <- withPrec 11 $ ppExpr e2
>          return (group (pk <+> pe1 <+> pe2))
> ppExpr (Throw e)
>   =  do
>        pk <- keyword' "throw"
>        parensPrec 10 $ do
>          pe <- withPrec 11 $ ppExpr e
>          return (group (pk <+> pe))
> ppExpr (Try e rs)
>   =  do
>        pkt <- keyword' "try"
>        pkc <- keyword' "catch"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        prs <- withPrec 0 $ mapM ppRule rs
>        return (group (pkt <> block 2 (sp <> pe) <+> pkc <+>
>                              block 2 (intersperse (sp <> text " | ") prs) <+> pke))
> ppExpr (Call f e)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        return (group (pretty f <+> pe))
> ppExpr (While e1 e2)
>   =  do
>        pkw <- keyword "while"
>        pkd <- keyword' "do"
>        pke <- keyword' "end"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        return (group (pkw <> 
>                       block 2 (pe1 <> sp <> pkd <+> pe2) <+> pke))

> ppExpr Empty
>   =  return (text "")
> ppExpr (Val p e)
>   =  do
>        pk <- keyword' "val"
>        pp <- withPrec 0 $ ppPat p
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pk <+> block 4 (pp <+> text "=" <+> pe)))
> ppExpr (Comma e1 e2)
>   =  do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <+> pe2))
> ppExpr (Lookup e x)
>   =  do
>        pe <- withPrec 11 $ ppExpr e
>        return (group (pe <> text "$" <> text x))

> ppExpr (Class e)
>   =  do
>        pkc <- keyword' "class"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pkc <+> pe <+> pke))
> ppExpr (Object e)
>   =  do
>        pko <- keyword' "object"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pko <+> pe <+> pke))
> ppExpr (New e)
>   =  do
>        pk <- keyword' "new"
>        parensPrec 10 $ do
>          pe <- withPrec 11 $ ppExpr e
>          return (group (pk <+> pe))
> ppExpr (Append e1 e2)
>   =  do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (group (pe1 <+> pe2))
> ppExpr (Method x e)
>   =  do
>        pk <- keyword' "method"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pk <+> text x <> text " =" <+> pe))
> ppExpr (Invoke e1 l e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 11 $ ppExpr e2
>        return (pe1 <> text "." <> ppLab l <> text "~" <> pe2)
> ppExpr (Inherit e)
>   =  do
>        pk <- keyword' "inherit"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pk <+> pe))
> ppExpr (InheritIn x e1 e2)
>   =  do
>        pkl <- keyword "inherit"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        return (group (pkl <> block 8 (text x <> text " =" <+> pe1) <> 
>                       block 1 (sp <> pki <> block 3 pe2) <+> pke))
> ppExpr (Include e)
>   =  do
>        pk <- keyword' "include"
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pk <+> pe))
> ppExpr (IncludeIn x e1 e2)
>   =  do
>        pkl <- keyword "include"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        return (group (pkl <> block 8 (text x <> text " =" <+> pe1) <> 
>                       block 1 (sp <> pki <> block 3 pe2) <+> pke))

> ppExpr (ArrFun e1 x e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        alignkey "array" (brackets pe1 <+> text x <> text " =>" <+> pe2)
> ppExpr (ArrList es)
>   =  ppList ppExpr es
> ppExpr (Subscript e1 e2)
>   =  parensPrec 10 $ do
>        pe1 <- withPrec 11 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        return (group (pe1 <> text "." <> brackets pe2))

 ppExpr e -- catch-all (vorlufig)
   =  pretty e

> ppRule :: (Pat, Expr) -> P Doc
> ppRule (p, e)
>   =  do
>        pp <- withPrec 0 $ ppPat p
>        pe <- withPrec 0 $ ppExpr e
>        return (group (pp <> text " => " <> pe))

> alignkey s d
>   =  do
>        k <- keyword s
>        return (block (length s + 1) (k <> d))

> align s d = block (length s) (text s <> d)

> ppList :: (a -> P Doc) -> [a] -> P Doc
> ppList pp as
>   =  fmap (brackets . intersperse (char ',' <> sp)) (mapM pp as)

> infixl $$
> ($$) = App

Records.

> data Row a
>   =  Named     String a
>   |  Numbered  a
>      deriving (Show)

> instance Eq (Row a) where
>   Named s1 _  == Named s2 _  =  s1 == s2
>   Numbered _  == Numbered _  =  True
>   _           == _           =  False
>
> instance Ord (Row a) where
>   compare (Named s1 _) (Named s2 _)  =  compare s1 s2
>   compare (Named _ _ ) (Numbered _)  =  LT
>   compare (Numbered _) (Named _ _ )  =  GT
>   compare (Numbered _) (Numbered _)  =  EQ

> ppRow pp (Named s a)
>   =  fmap (align (s ++ " = ")) (pp a)
> ppRow pp (Numbered a)
>   =  pp a

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Patterns}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Pat
>   =  PatFail                                          -- 0
>   |  PatSucceed                                       -- 1
>   |  PatVar     Var
>   |  PatCon     Con Pat
>   |  PatType                                          -- type abstraction marker
>   |  PatRecord  [Row Pat]
>   |  PatBoth    Pat Pat                               -- &
>   |  PatAlt     Pat Pat                               -- |
>      deriving (Show)

> patUnit       =  PatRecord []
> patRecord rs  =  PatRecord (sort rs)

NB. `sort' must be a stable sorting algorithm, so that the relative
order of numbered fields is not changed.

> ppPat :: Pat -> P Doc
> ppPat PatFail
>   =  keyword' "fail"
> ppPat PatSucceed
>   =  return (text "_")
> ppPat (PatVar x)
>   =  return (text x)
> ppPat (PatCon c p)
>   =  parensPrec 10 $ do
>        pc <- ppCon c
>        pp <- withPrec 11 $ ppPat p
>        return (pc <+> pp)
> ppPat PatType
>   =  do
>        pc <- ppCol astype
>        return (pc (text "<>"))
> ppPat (PatRecord rs)
>   =  do
>        prs <- withPrec 0 $ mapM (ppRow ppPat) rs
>        return (parens (group (intersperse (char ',' <> sp) prs)))
> ppPat (PatAlt p1 p2)
>   =  parensPrec 10 $ do
>        pp1 <- withPrec 11 $ ppPat p1
>        pp2 <- withPrec 11 $ ppPat p2
>        return (group (pp1 <> text " " <> block 2 (text "|" <+> pp2)))
> ppPat (PatBoth p1 p2)
>   =  parensPrec 10 $ do
>        pp1 <- withPrec 11 $ ppPat p1
>        pp2 <- withPrec 11 $ ppPat p2
>        return (group (pp1 <> text " " <> block 2 (text "&" <+> pp2)))

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{AST to core}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> core :: AS.Expr -> Expr
> core (AS.Bool _ b) 
>   =  Bool b
> core (AS.If e1 e2 e3) 
>   =  If (core e1) (core e2) (core e3)
> core (AS.UnOp op e) 
>   =  UnOp op (core e)
> core (AS.BinOp e1 Or e2)
>   =  If (core e1) (Bool True) (core e2)
> core (AS.BinOp e1 And e2)
>   =  If (core e1) (core e2) (Bool False)
> core (AS.BinOp e1 op e2) 
>   =  BinOp (core e1) op (core e2)
> core (AS.Nat _ n)
>   =  Nat n
> core (AS.String _ s) 
>   =  String s
> core (AS.Var (_,x)) 
>   =  Var x
> core (AS.Let ds e) 
>   =  Let (coreDecls ds) (core e)
> core (AS.Fun xs e)
>   =  coreAbs xs (core e)
> core (AS.RecFun (_,f) xs t e)
>   =  Rec (PatVar f) (coreAbs xs (core e))
> core (AS.App (AS.QualCon _ (_,c)) [AS.ActExpr e])         -- unary constructor
>   =  Con c (core e)
> core (AS.App (AS.Con (_,c)) [AS.ActExpr e])               -- unary constructor
>   =  Con c (core e)
> core (AS.App e es) 
>   =  coreApp (core e) es
> core (AS.Rec p e) 
>   =  Rec (corePat p) (core e)
> core (AS.Record rs) 
>   =  Record (map (coreRow core) rs)
> core (AS.Con (_,c))
>   =  Con c unit                                           -- nullary constructor

> core (AS.Sel e l) = Sel (core e) l

> core (AS.Seq e1 e2) = Seq (core e1) (core e2)
> core (AS.Assign e1 e2) = Assign (core e1) (core e2)
> core (AS.Ref e) = Ref (core e)
> core (AS.Deref e) = Deref (core e)

> core (AS.Continue e1 e2)
>   =  Continue (core e1) (core e2)
> core (AS.LetCC (_,x) t e)
>   =  LetCC x (core e)

> core (AS.Case e rs) = Case (core e) (map coreRule rs)
> core (AS.Throw e) = Throw (core e)
> core (AS.Try e rs) = Try (core e) (map coreRule rs)
> -- core (AS.Call f e) = Call f (core e)
> core (AS.While e1 e2) = While (core e1) (core e2)
> core (AS.QualVar e (_,x)) = Lookup (coreStr e) x
> core (AS.QualCon e (_,c)) = Con c (unit)                  -- nullary constructor !! e wird ignoriert

> core (AS.Class ss)
>   =  Class (coreSlots ss)
> core (AS.Object ss)
>   =  Object (coreSlots ss)
> core (AS.New e) 
>   =  New (core e)
> core (AS.Invoke e1 l e2) 
>   = Invoke (core e1) l (core e2)

> core (AS.ArrFun e1 (_,x) e2) 
>   =  ArrFun (core e1) x (core e2)
> core (AS.ArrList es) 
>   =  ArrList (map core es)
> core (AS.Subscript e1 e2) 
>   =  Subscript (core e1) (core e2)
> core (AS.Typed e _t)
>   =  core e

> core e = error (show e)

> coreRule (p, e)
>   =  (corePat p, core e)

> corePat :: AS.Pat -> Pat
> corePat AS.PatSucceed
>   =  PatSucceed
> corePat (AS.PatVar (_,x)) 
>   =  PatVar x
> corePat (AS.PatQualCon _ c p)
>   =  corePat (AS.PatCon c p)  -- ignore qualifier for evaluation
> corePat (AS.PatCon (_,c) Nothing) 
>   =  PatCon c patUnit
> corePat (AS.PatCon (_,c) (Just p)) 
>   =  PatCon c (corePat p)
> corePat (AS.PatTyped p _t) 
>   =  corePat p
> corePat (AS.PatRecord rs)
>   =  patRecord (map (coreRow corePat) rs)
> corePat (AS.PatAlt p1 p2) 
>   =  PatAlt (corePat p1) (corePat p2)
> corePat (AS.PatBoth p1 p2) 
>   =  PatBoth (corePat p1) (corePat p2)

> coreAbs [] e 
>   =  e

> coreAbs (AS.ForPat p : xs) e
>   =  Fun (corePat p) (coreAbs xs e)
> coreAbs (AS.ForTypePat [] : xs) e
>   =  coreAbs xs e
> coreAbs (AS.ForTypePat (a : as) : xs) e
>   =  Fun PatType (coreAbs (AS.ForTypePat as : xs) e)

Die Typabstraktion `\ <a> => e' wird nach `\ () => e' berfhrt; die
Typapplikation `e <t>' entsprechend nach `e ()'. Das ist notwendig, um
die Auswertung zu verzgern (\ <a> => ref (raise Undefined)), siehe
auch Pierce, Types and Programming Languages, Abschnitt 23.7.
Sprachen mit implizitem Polymorphismus mssen z.B. die Form von
polymorphen Ausdrcken einschrnken (value restriction in Standard ML).

> coreApp e [] 
>   =  e
> coreApp e1 (AS.ActExpr e2 : es) 
>   =  coreApp (App e1 (core e2)) es
> coreApp e1 (AS.ActType [] : es) 
>   =  coreApp e1 es
> coreApp e1 (AS.ActType (t : ts) : es) 
>   =  coreApp (App e1 Type) (AS.ActType ts : es)

Translation of declarations.

> coreDecls = foldr comma Empty . map coreDecl

> coreDecl (AS.Val p e)
>   =  Val (corePat p) (core e)
> coreDecl (AS.Function [((_, f), xs, Nothing, e)])  -- single non-recursive function
>   =  Val (PatVar f) (coreAbs xs (core e))
> coreDecl (AS.Function [((_, f), xs, Just t, e)])   -- single recursive function
>   =  Val (PatVar f) (Rec (PatVar f) (coreAbs xs (core e)))
> coreDecl (AS.Function fs)                          -- group of mutually recursive functions
>   =  Val p (Rec p e)
>      where p = patRecord [ Named f (PatVar f)            | ((_, f), xs, t, e) <- fs ]
>            e = Record    [ Named f (coreAbs xs (core e)) | ((_, f), xs, t, e) <- fs ]
> coreDecl (AS.Types _)
>   =  Empty
> coreDecl (AS.Constr _ _)
>   =  Empty
> coreDecl (AS.Exception _)
>   =  Empty
> coreDecl (AS.Module (_, x1) fs _t e1)
>   =  Val (PatVar x1) (coreStr (AS.Functor fs e1))
> coreDecl (AS.Open e)
> --  =  Open (coreStr e)
>   =  coreStr e
> coreDecl (AS.OpenAbstract e)
> --  =  Open (coreStr e)
>   =  coreStr e
> coreDecl (AS.Local ds1 ds2)
>   =  Let (coreDecls ds1) (coreDecls ds2)
> coreDecl (AS.Interface _ _ _)
>   =  Empty

> coreRow c (AS.Field Nothing e)
>   =  Numbered (c e)
> coreRow c (AS.Field (Just (AlNumLab l)) e)
>   =  Named l (c e)

> coreStr (AS.StrVar (_,x))
>   =  Var x
> coreStr (AS.StrSel e (_,x))
>   =  Lookup (coreStr e) x
> coreStr (AS.Struct ds)
> --  =  Struct (coreDecls ds)
>   =  coreDecls ds
> coreStr (AS.RecStruct (_, x) _ ds)
>   =  Rec (PatVar x) (coreStr (AS.Struct ds))
> coreStr (AS.Functor fs e)
>   =  coreStrAbs fs (coreStr e)
> coreStr (AS.StrApp e as)
>   =  coreStrApp (coreStr e) as
> coreStr (AS.StrTyped e _t)
>   =  coreStr e

> coreStrApp e []
>   =  e
> coreStrApp e1 (AS.StrActStrExpr e2 : es)
>   =  coreStrApp (App e1 (coreStr e2)) es
> coreStrApp e1 (AS.StrActType [] : es)
>   =  coreStrApp e1 es
> coreStrApp e1 (AS.StrActType (t : ts) : es) 
>   =  coreStrApp (App e1 Type) (AS.StrActType ts : es)

> coreStrAbs [] e
>   =  e
> coreStrAbs (AS.StrForStrPat (_,x) t : xs) e
>   =  Fun (PatVar x) (coreStrAbs xs e)
> coreStrAbs (AS.StrForTypePat [] : xs) e
>   =  coreStrAbs xs e
> coreStrAbs (AS.StrForTypePat (a : as) : xs) e
>   =  Fun PatType (coreStrAbs (AS.StrForTypePat as : xs) e)

 coreCls (AS.ClsVar x)
   =  Var x
 coreCls (AS.Object ds)
   =  Delay (Push (coreLet ds Object))
 coreCls (AS.ClsApp e as)
   =  coreApp (coreCls e) as
 coreCls (AS.ClsFun xs e)
   =  coreAbs xs (coreCls e)
 coreCls (AS.ClsLet ds e)
   =  coreLet ds (coreCls e)

> coreSlots = foldr append Empty . map coreSlot

> coreSlot (AS.SlMethod (_,m) fs t e)
>  =  Method m (coreAbs fs (core e))
> coreSlot (AS.SlLocal ds ss)
>  =  Let (coreDecls ds) (coreSlots ss)
> coreSlot (AS.SlInherit e)
>  =  Inherit (core e)
> coreSlot (AS.SlInheritIn (_,x) e ss)
>  =  InheritIn x (core e) (coreSlots ss)
> coreSlot (AS.SlInclude e)
>  =  Include (core e)
> coreSlot (AS.SlIncludeIn (_,x) e ss)
>  =  IncludeIn x (core e) (coreSlots ss)

> expr2core :: AS.Expr -> Pass Expr
> expr2core e
>   =  let c = core e in
>      Succeed c  [Subtitle (text "abstract syntax tree (expr) -> core"),
>                  Debug (text "core"),
>                  Debug (pretty c)]

> decls2core :: [AS.Decl] -> Pass Expr
> decls2core ds
>   =  let c = coreDecls ds in
>      Succeed c  [Subtitle (text "abstract syntax tree (module) -> core"),
>                  Debug (text "core"),
>                  Debug (pretty c)]

> free :: Expr -> Set Var
> free e =
>   case e of
>     If e1 e2 e3 -> unions (map free [e1,e2,e3])
>     UnOp _ e    -> free e
>     BinOp e1 _ e2 -> union (free e1) (free e2)
>     Var x       -> singleton x

TODO: anpassen:
     Let p e1 e2 -> union (free e1) (difference (free e2) (freePat p))

>     Fun p e     -> difference (free e) (freePat p)
>     App e1 e2   -> union (free e1) (free e2)
>     Rec p e     -> difference (free e) (freePat p)
>     Record rs   -> unions (map (freeRow free) rs)
>     Sel e _     -> free e
>     Con c e     -> {- union (singleton c) -} (free e)  -- TODO
>     Case e cs   -> unions (free e : map (\ (p, e) -> difference (free e) (freePat p)) cs)
>     Seq e1 e2   -> union (free e1) (free e2)
>     Assign e1 e2 -> union (free e1) (free e2)
>     Ref e       -> free e
>     Deref e     -> free e
>     LetCC x e   -> difference (free e) (singleton x)
>     Continue e1 e2 -> union (free e1) (free e2)
>     Throw e     -> free e
>     Try e cs    -> unions (free e : map (\ (p, e) -> difference (free e) (freePat p)) cs)
>     Call p e    -> free e
>     While e1 e2 -> union (free e1) (free e2)

TODO: Regeln fr 1st-class environments

>     Lookup e _  -> free e  -- ?
>     Class e     -> free e
>     Object e    -> free e
>     New e       -> free e
>     Append e1 e2 -> union (free e1) (free e2)
>     Method _ e  -> free e  -- ?
>     Invoke e1 _ e2 -> union (free e1) (free e2)
>     Inherit e   -> free e
>     InheritIn x e1 e2 -> union (free e1) (difference (free e2) (singleton x))
>     Include e   -> free e
>     IncludeIn x e1 e2 -> union (free e1) (difference (free e2) (singleton x))
>     ArrFun e1 x e2 -> union (free e1) (difference (free e2) (singleton x))
>     ArrList es  -> unions (map free es)
>     Subscript e1 e2 -> union (free e1) (free e2)
>     _           -> S.empty

> freePat :: Pat -> Set Var
> freePat p =
>   case p of
>     PatVar x    -> singleton x
>     PatCon x p  -> freePat p  -- TODO: determine constructors for patterns
>     PatRecord rs -> unions (map (freeRow freePat) rs)
>     PatBoth p1 p2 -> union (freePat p1) (freePat p2)
>     PatAlt p1 p2 -> union (freePat p1) (freePat p2)
>     _           -> S.empty

> freeRow :: (a -> Set Var) -> Row a -> Set Var
> freeRow f r =
>   case r of
>     Named _ e   -> f e
>     Numbered e  -> f e
